"""
Django admin declarations
"""

# Django
from django.contrib import admin, messages
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

# AA SRP
from aasrp.form import SettingAdminForm
from aasrp.models import FleetType, RequestComment, Setting, SrpLink, SrpRequest


def custom_filter(title):
    """
    Custom filter for model properties
    :param title:
    :return:
    """

    class Wrapper(admin.FieldListFilter):
        """
        Custom_filter wrapper
        """

        def expected_parameters(self):
            """
            Expected parameters

            :return:
            :rtype:
            """

            pass  # pylint: disable=unnecessary-pass

        def choices(self, changelist):
            """
            Choices

            :param changelist:
            :type changelist:
            :return:
            :rtype:
            """

            pass  # pylint: disable=unnecessary-pass

        def __new__(cls, *args, **kwargs):
            """
            New instance

            :param args:
            :type args:
            :param kwargs:
            :type kwargs:
            """

            instance = admin.FieldListFilter.create(*args, **kwargs)
            instance.title = title

            return instance

    return Wrapper


class SingletonModelAdmin(admin.ModelAdmin):
    """
    Prevents Django admin users deleting the singleton or adding extra rows.
    """

    actions = None  # Removes the default delete action.

    def has_add_permission(self, request):  # pylint: disable=unused-argument
        """
        Has "add" permissions

        :param request:
        :type request:
        :return:
        :rtype:
        """

        return self.model.objects.all().count() == 0

    def has_change_permission(
        self, request, obj=None  # pylint: disable=unused-argument
    ):
        """
        Has "change" permissions

        :param request:
        :type request:
        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return True

    def has_delete_permission(
        self, request, obj=None  # pylint: disable=unused-argument
    ):
        """
        Has "delete" permissions

        :param request:
        :type request:
        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return False


@admin.register(SrpLink)
class SrpLinkAdmin(admin.ModelAdmin):
    """
    SrpLinkAdmin
    """

    list_display = (
        "srp_code",
        "fleet_time",
        "_creator",
        "srp_name",
        "srp_status",
        "fleet_doctrine",
    )
    ordering = ("fleet_time",)

    list_filter = ("creator", "srp_status", "fleet_doctrine")

    search_fields = ("srp_code", "fleet_doctrine", "srp_name")

    @classmethod
    @admin.display(description=_("Creator"), ordering="creator")
    def _creator(cls, obj):
        """
        Display the creator name

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        creator_name = obj.creator

        if obj.creator.profile.main_character:
            creator_name = obj.creator.profile.main_character.character_name

        return creator_name


@admin.register(SrpRequest)
class SrpRequestAdmin(admin.ModelAdmin):
    """
    SrpRequestAdmin
    """

    list_display = (
        "request_code",
        "_creator",
        "character",
        "srp_link",
        "post_time",
        "ship",
        "loss_amount",
        "payout_amount",
        "killboard_link",
        "request_status",
    )
    ordering = ("post_time",)

    list_filter = ("creator", "character", "request_status")

    search_fields = (
        "request_code",
        "character__character_name",
        "ship__name",
        "srp_link__srp_code",
    )

    @classmethod
    @admin.display(description=_("Requestor"), ordering="creator")
    def _creator(cls, obj):
        """
        Display the creator name

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        creator_name = obj.creator

        if obj.creator.profile.main_character:
            creator_name = obj.creator.profile.main_character.character_name

        return creator_name


@admin.register(RequestComment)
class RequestCommentAdmin(admin.ModelAdmin):
    """
    RequestCommentAdmin
    """

    list_display = ("srp_request", "comment_type", "creator")
    ordering = ("srp_request",)
    list_filter = ("comment_type",)


@admin.register(FleetType)
class FleetTypeAdmin(admin.ModelAdmin):
    """
    FleetTypeAdmin
    """

    list_display = ("id", "_name", "_is_enabled")
    list_filter = ("is_enabled",)
    ordering = ("name",)

    @admin.display(description=_("Fleet type"), ordering="name")
    def _name(self, obj):
        """
        Rewrite name

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return obj.name

    @admin.display(description=_("Is enabled"), boolean=True, ordering="is_enabled")
    def _is_enabled(self, obj):
        """
        Rewrite is_enabled

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return obj.is_enabled

    actions = ("activate", "deactivate")

    @admin.action(description=_("Activate selected fleet types"))
    def activate(self, request, queryset):
        """
        Mark fleet type as active

        :param request:
        :type request:
        :param queryset:
        :type queryset:
        :return:
        :rtype:
        """

        notifications_count = 0
        failed = 0

        for obj in queryset:
            try:
                obj.is_enabled = True
                obj.save()

                notifications_count += 1
            except Exception:  # pylint: disable=broad-exception-caught
                failed += 1

        if failed:
            messages.error(
                request=request,
                message=ngettext(
                    singular="Failed to activate {failed} fleet type",
                    plural="Failed to activate {failed} fleet types",
                    number=failed,
                ).format(failed=failed),
            )

        if queryset.count() - failed > 0:
            messages.success(
                request=request,
                message=ngettext(
                    singular="Activated {notifications_count} fleet type",
                    plural="Activated {notifications_count} fleet types",
                    number=notifications_count,
                ).format(notifications_count=notifications_count),
            )

    @admin.action(description=_("Deactivate selected fleet types"))
    def deactivate(self, request, queryset):
        """
        Mark fleet type as inactive

        :param request:
        :type request:
        :param queryset:
        :type queryset:
        :return:
        :rtype:
        """

        notifications_count = 0
        failed = 0

        for obj in queryset:
            try:
                obj.is_enabled = False
                obj.save()

                notifications_count += 1
            except Exception:  # pylint: disable=broad-exception-caught
                failed += 1

        if failed:
            messages.error(
                request=request,
                message=ngettext(
                    singular="Failed to deactivate {failed} fleet type",
                    plural="Failed to deactivate {failed} fleet types",
                    number=failed,
                ).format(failed=failed),
            )

        if queryset.count() - failed > 0:
            messages.success(
                request=request,
                message=ngettext(
                    singular="Deactivated {notifications_count} fleet type",
                    plural="Deactivated {notifications_count} fleet types",
                    number=notifications_count,
                ).format(notifications_count=notifications_count),
            )


@admin.register(Setting)
class SettingAdmin(SingletonModelAdmin):
    """
    Setting Admin
    """

    form = SettingAdminForm



"""
Our Models
"""

# Django
from django.contrib.auth.models import User
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_main_character_name_from_user

# Alliance Auth (External Libs)
from eveuniverse.models import EveType

# AA SRP
from aasrp.managers import SettingManager


def get_sentinel_user():
    """
    Get or create sentinel user
    :return:
    """

    return User.objects.get_or_create(username="deleted")[0]


class SingletonModel(models.Model):
    """
    SingletonModel
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Model meta definitions
        """

        abstract = True

    def save(self, *args, **kwargs):
        """
        Save action

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if self.__class__.objects.count():
            self.pk = self.__class__.objects.first().pk

        super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """
        Delete action

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        pass  # pylint: disable=unnecessary-pass


class AaSrp(models.Model):
    """
    Meta model for app permissions
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        General definitions
        """

        verbose_name = "AA-SRP"
        managed = False
        default_permissions = ()
        permissions = (
            # Can open the SRP app and submit SRP requests
            ("basic_access", "Can access the AA-SRP module"),
            # Can create SRP links
            ("create_srp", "Can create new SRP links"),
            # Can manage the complete SRP module
            ("manage_srp", "Can manage SRP"),
            # Can manage SRP requests only
            ("manage_srp_requests", "Can manage SRP requests"),
        )


class FleetType(models.Model):
    """
    FleetType
    """

    id = models.AutoField(primary_key=True)

    name = models.CharField(
        max_length=254,
        help_text=_("Descriptive name of your fleet type"),
        verbose_name=_("Name"),
    )

    is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=_("Whether this fleet type is active or not"),
        verbose_name=_("Is enabled"),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        AFatLinkType :: Meta
        """

        default_permissions = ()
        verbose_name = _("Fleet type")
        verbose_name_plural = _("Fleet types")

    def __str__(self) -> str:
        """
        Return the objects string name

        :return:
        :rtype:
        """

        return str(self.name)


class SrpLink(models.Model):
    """
    SRP link model
    """

    class Status(models.TextChoices):
        """
        Choices for SRP status
        """

        ACTIVE = "Active", _("Active")
        CLOSED = "Closed", _("Closed")
        COMPLETED = "Completed", _("Completed")

    srp_name = models.CharField(max_length=254, default="", verbose_name=_("SRP name"))
    srp_status = models.CharField(
        max_length=9,
        choices=Status.choices,
        default=Status.ACTIVE,
        verbose_name=_("SRP status"),
    )
    srp_code = models.CharField(max_length=16, default="", verbose_name=_("SRP code"))
    fleet_commander = models.ForeignKey(
        EveCharacter,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET_NULL,
        verbose_name=_("Fleet commander"),
    )
    fleet_doctrine = models.CharField(
        max_length=254, default="", verbose_name=_("Doctrine")
    )

    fleet_type = models.ForeignKey(
        FleetType,
        related_name="+",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        default=None,
        help_text=_("The SRP link fleet type, if it's set"),
        verbose_name=_("Fleet type"),
    )

    fleet_time = models.DateTimeField(verbose_name=_("Fleet time"))
    aar_link = models.CharField(
        max_length=254, blank=True, default="", verbose_name=_("AAR link")
    )

    creator = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(value=get_sentinel_user),
        help_text=_("Who created the SRP link?"),
        verbose_name=_("Creator"),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("SRP link")
        verbose_name_plural = _("SRP links")

    def __str__(self) -> str:
        """
        Return the objects string name

        :return:
        :rtype:
        """

        return str(self.srp_name)

    @property
    def total_cost(self):
        """
        Total cost for this SRP link

        :return:
        :rtype:
        """

        return sum(
            int(r.payout_amount)
            for r in self.srp_requests.filter(request_status=SrpRequest.Status.APPROVED)
        )

    @property
    def total_requests(self):
        """
        Number of total SRP requests

        :return:
        :rtype:
        """

        return self.srp_requests.count()

    @property
    def pending_requests(self):
        """
        Number of pending SRP requests

        :return:
        :rtype:
        """

        return self.srp_requests.filter(
            request_status=SrpRequest.Status.PENDING
        ).count()

    @property
    def approved_requests(self):
        """
        Number of approved SRP requests

        :return:
        :rtype:
        """

        return self.srp_requests.filter(
            request_status=SrpRequest.Status.APPROVED
        ).count()

    @property
    def rejected_requests(self):
        """
        Number of rejected SRP requests

        :return:
        :rtype:
        """

        return self.srp_requests.filter(
            request_status=SrpRequest.Status.REJECTED
        ).count()

    @property
    def requests(self):
        """
        All SRP requests

        :return:
        :rtype:
        """

        return self.srp_requests.all()


class SrpRequest(models.Model):
    """
    SRP Request model
    """

    class Status(models.TextChoices):
        """
        Choices for SRP Request status
        """

        PENDING = "Pending", _("Pending")
        APPROVED = "Approved", _("Approved")
        REJECTED = "Rejected", _("Rejected")

    request_code = models.CharField(
        max_length=254, default="", verbose_name=_("Request code")
    )
    creator = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(value=get_sentinel_user),
        help_text=_("Who created the SRP request?"),
        verbose_name=_("Creator"),
    )
    character = models.ForeignKey(
        EveCharacter,
        related_name="+",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Character"),
    )
    ship_name = models.CharField(
        max_length=254, default="", verbose_name=_("Ship type")
    )
    ship = models.ForeignKey(
        EveType,
        related_name="srp_requests",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET_NULL,
        verbose_name=_("Ship type"),
    )
    killboard_link = models.CharField(
        max_length=254, default="", verbose_name=_("Killboard link")
    )
    additional_info = models.TextField(
        blank=True, default="", verbose_name=_("Additional information")
    )
    request_status = models.CharField(
        max_length=8,
        choices=Status.choices,
        default=Status.PENDING,
        verbose_name=_("Request status"),
    )
    payout_amount = models.BigIntegerField(default=0, verbose_name=_("Payout amount"))
    srp_link = models.ForeignKey(
        SrpLink,
        related_name="srp_requests",
        on_delete=models.CASCADE,
        verbose_name=_("SRP link"),
    )
    loss_amount = models.BigIntegerField(default=0, verbose_name=_("Loss amount"))
    post_time = models.DateTimeField(
        default=timezone.now, verbose_name=_("Request time")
    )
    reject_info = models.TextField(
        blank=True, default="", verbose_name=_("Reject reason")
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Request")
        verbose_name_plural = _("Requests")

    def __str__(self):
        """
        Return the objects string name

        :return:
        :rtype:
        """

        character_name = self.character.character_name
        user_name = get_main_character_name_from_user(self.creator)
        ship = self.ship.name
        request_code = self.request_code

        return _(
            "{character_name} ({user_name}) SRP request for: {ship} ({request_code})"
        ).format(
            character_name=character_name,
            user_name=user_name,
            ship=ship,
            request_code=request_code,
        )


class Insurance(models.Model):
    """
    Insurance Model
    """

    srp_request = models.ForeignKey(
        SrpRequest,
        on_delete=models.CASCADE,
        related_name="insurance",
        verbose_name=_("SRP request"),
    )
    insurance_level = models.CharField(
        max_length=254, default="", verbose_name=_("Insurance level")
    )
    insurance_cost = models.FloatField(verbose_name=_("Insurance cost"))
    insurance_payout = models.FloatField(verbose_name=_("Insurance payout"))

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Ship insurance")
        verbose_name_plural = _("Ship insurances")


class RequestComment(models.Model):
    """
    SRP Request Comments model
    """

    class Type(models.TextChoices):
        """
        Choices for comment types
        """

        COMMENT = "Comment", _("Comment")
        REQUEST_ADDED = "Request Added", _("SRP request added")
        REQUEST_INFO = "Request Information", _("Additional information")
        REJECT_REASON = "Reject Reason", _("Reject reason")
        STATUS_CHANGE = "Status Changed", _("Status changed")
        REVISER_COMMENT = "Reviser Comment", _("Reviser comment")

    comment = models.TextField(blank=True, default="", verbose_name=_("Comment"))

    comment_type = models.CharField(
        max_length=19,
        choices=Type.choices,
        default=Type.COMMENT,
        verbose_name=_("Comment type"),
    )

    creator = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(value=get_sentinel_user),
        verbose_name=_("Creator"),
    )

    srp_request = models.ForeignKey(
        SrpRequest,
        related_name="srp_request_comments",
        null=True,
        blank=True,
        default=None,
        on_delete=models.CASCADE,
        verbose_name=_("SRP request"),
    )

    comment_time = models.DateTimeField(
        default=timezone.now,
        null=True,
        blank=True,
        # Translators: This is the time when the comment was made
        verbose_name=_("Comment time"),
    )

    new_status = models.CharField(
        max_length=8,
        choices=SrpRequest.Status.choices,
        blank=True,
        default="",
        # Translators: New SRP request status that might have been set
        verbose_name=_("New SRP request status"),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Comment")
        verbose_name_plural = _("Comments")


class UserSetting(models.Model):
    """
    User settings
    """

    user = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(value=get_sentinel_user),
        verbose_name=_("User"),
    )

    disable_notifications = models.BooleanField(
        default=False, verbose_name=_("Disable notifications")
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("User settings")
        verbose_name_plural = _("User settings")


class Setting(SingletonModel):
    """
    SRP module settings
    """

    class Field(models.TextChoices):
        """
        Choices for Setting.Field
        """

        SRP_TEAM_DISCORD_CHANNEL_ID = "srp_team_discord_channel_id", _(
            "SRP team Discord channel ID"
        )

        LOSS_VALUE_SOURCE = "loss_value_source", _("Loss value source")

    class LossValueSource(models.TextChoices):
        """
        Choices for Setting.LossValueSource
        """

        FITTED_VALUE = "fittedValue", _("Fitted value (Ship and Fitting)")
        TOTAL_VALUE = "totalValue", _("Total value (Ship, Fitting and Cargo)")

    srp_team_discord_channel_id = models.PositiveBigIntegerField(
        blank=True,
        default=None,
        null=True,
        verbose_name=Field.SRP_TEAM_DISCORD_CHANNEL_ID.label,  # pylint: disable=no-member
    )

    loss_value_source = models.CharField(
        choices=LossValueSource.choices,
        default=LossValueSource.TOTAL_VALUE,
        help_text=_(
            "The source for the loss value of a killmail. "
            "Fitted value is the value of the ship and its fitting. "
            "Total value is the value of the ship, its fitting and the cargo. "
            "(Default: Total value)"
        ),
        max_length=11,
        verbose_name=Field.LOSS_VALUE_SOURCE.label,  # pylint: disable=no-member
    )

    objects = SettingManager()

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Setting")
        verbose_name_plural = _("Settings")

    def __str__(self) -> str:
        """
        Return the objects string name

        :return:
        :rtype:
        """

        return str(_("AA-SRP settings"))



"""
Test auth_hooks
"""

# Standard Library
from http import HTTPStatus
from unittest.mock import Mock, patch

# Django
from django.test import TestCase
from django.urls import reverse

# AA SRP
from aasrp import __title__
from aasrp.auth_hooks import AaSrpMenuItem, register_menu
from aasrp.tests.utils import create_fake_user


class TestMenuItemHtml(TestCase):
    """
    Test the html of the menu item
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and users
        """

        super().setUpClass()

        # User cannot access
        cls.user_1001 = create_fake_user(
            character_id=1001, character_name="Peter Parker"
        )

        # User can access
        cls.user_1002 = create_fake_user(
            character_id=1002,
            character_name="Bruce Wayne",
            permissions=["aasrp.basic_access"],
        )

        cls.html_menu = f"""
            <li class="d-flex flex-wrap m-2 p-2 pt-0 pb-0 mt-0 mb-0 me-0 pe-0">
                <i class="nav-link fa-regular fa-money-bill-1 fa-fw align-self-center me-3 "></i>
                <a class="nav-link flex-fill align-self-center me-auto" href="{reverse('aasrp:srp_links')}">
                    {__title__}
                </a>
            </li>
        """

    def test_render_hook_success(self):
        """
        Test should show the link to the app in the navigation to user with access

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_1002)

        response = self.client.get(path=reverse(viewname="authentication:dashboard"))

        self.assertEqual(first=response.status_code, second=HTTPStatus.OK)
        self.assertContains(response=response, text=self.html_menu, html=True)

    def test_render_hook_fail(self):
        """
        Test should not show the link to the app in the
        navigation to user without access

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_1001)

        response = self.client.get(path=reverse(viewname="authentication:dashboard"))

        self.assertEqual(first=response.status_code, second=HTTPStatus.OK)
        self.assertNotContains(response=response, text=self.html_menu, html=True)


class TestAaSrpMenuItem(TestCase):
    """
    Test the menu item
    """

    def setUp(self):
        """
        Set up the test

        :return:
        :rtype:
        """

        self.request = Mock()
        self.request.path = reverse(viewname="authentication:dashboard")
        self.menu_item = AaSrpMenuItem()

    def test_render_no_permission(self):
        """
        Test should return empty string if user has no permission

        :return:
        :rtype:
        """

        self.request.user.has_perm.return_value = False
        result = self.menu_item.render(self.request)

        self.assertEqual(result, "")

    @patch("aasrp.auth_hooks.SrpManager.pending_requests_count_for_user")
    def test_render_with_permission(self, mock_pending_requests):
        """
        Test should return html if a user has permission

        :param mock_pending_requests:
        :type mock_pending_requests:
        :return:
        :rtype:
        """

        self.request.user.has_perm.return_value = True
        mock_pending_requests.return_value = 5
        result = self.menu_item.render(self.request)

        self.assertIsInstance(result, str)


class TestRegisterMenu(TestCase):
    """
    Test register menu
    """

    def test_register_menu(self):
        """
        Test should return menu item

        :return:
        :rtype:
        """

        result = register_menu()

        self.assertIsInstance(result, AaSrpMenuItem)



"""
Helper for our tests
"""

# Standard Library
import re

# Third Party
from faker import Faker

# Django
from django.contrib.auth.models import User

# Alliance Auth
from allianceauth.tests.auth_utils import AuthUtils

# AA SRP
from aasrp.models import FleetType, Setting

fake = Faker()


def create_fake_user(
    character_id: int,
    character_name: str,
    corporation_id: int = None,
    corporation_name: str = None,
    corporation_ticker: str = None,
    permissions: list[str] = None,
    **kwargs,
) -> User:
    """
    Create a fake user incl. Main character and (optional) permissions.

    :param character_id:
    :type character_id:
    :param character_name:
    :type character_name:
    :param corporation_id:
    :type corporation_id:
    :param corporation_name:
    :type corporation_name:
    :param corporation_ticker:
    :type corporation_ticker:
    :param permissions:
    :type permissions:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    username = re.sub(pattern=r"[^\w\d@\.\+-]", repl="_", string=character_name)
    user = AuthUtils.create_user(username=username)

    if not corporation_id:
        corporation_id = 2001
        corporation_name = "Wayne Technologies Inc."
        corporation_ticker = "WTE"

    alliance_id = kwargs.get("alliance_id", 3001)
    alliance_name = (
        kwargs.get("alliance_name", "Wayne Enterprises")
        if alliance_id is not None
        else ""
    )

    AuthUtils.add_main_character_2(
        user=user,
        name=character_name,
        character_id=character_id,
        corp_id=corporation_id,
        corp_name=corporation_name,
        corp_ticker=corporation_ticker,
        alliance_id=alliance_id,
        alliance_name=alliance_name,
    )

    if permissions:
        perm_objs = [
            AuthUtils.get_permission_by_name(perm=perm) for perm in permissions
        ]
        user = AuthUtils.add_permissions_to_user(perms=perm_objs, user=user)

    return user


def get_or_create_fake_user(*args, **kwargs) -> User:
    """
    Same as create_fake_user but will not fail when user already exists.

    :param args:
    :type args:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    if len(args) > 1:
        character_name = args[1]
    elif "character_name" in kwargs:
        character_name = kwargs["character_name"]
    else:
        ValueError("character_name is not defined")

    username = character_name.replace("'", "").replace(" ", "_")

    try:
        return User.objects.get(username=username)
    except User.DoesNotExist:
        return create_fake_user(*args, **kwargs)


def create_fleettype(**kwargs) -> FleetType:
    """
    Create a fleet type

    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    if "name" not in kwargs:
        kwargs["name"] = fake.name()

    return FleetType.objects.create(**kwargs)


def create_setting(**kwargs) -> Setting:
    """
    Create setting

    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    return Setting.objects.create(**kwargs)



"""
SRP Manager
"""

# pylint: disable=cyclic-import

# Third Party
import requests

# Django
from django.contrib.auth.models import User
from django.db import models

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA SRP
from aasrp import __title__
from aasrp.constants import USERAGENT, ZKILLBOARD_API_URL
from aasrp.providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class SrpManager:
    """
    AaSrpManager
    """

    @staticmethod
    def get_kill_id(killboard_link: str):
        """
        Get killmail ID from zKillboard link

        :param killboard_link:
        :type killboard_link:
        :return:
        :rtype:
        """

        num_set = "0123456789"
        kill_id = "".join(c for c in killboard_link if c in num_set)

        return kill_id

    @staticmethod
    def get_kill_data(kill_id: str):  # pylint: disable=too-many-locals
        """
        Get kill data from zKillboard

        :param kill_id:
        :type kill_id:
        :return:
        :rtype:
        """

        url = f"{ZKILLBOARD_API_URL}killID/{kill_id}/"
        headers = {"User-Agent": USERAGENT, "Content-Type": "application/json"}
        request_result = requests.get(url=url, headers=headers, timeout=5)

        try:
            request_result.raise_for_status()
        except requests.HTTPError as exc:
            error_str = str(exc)

            logger.warning(
                msg=f"Unable to get kill mail details from zKillboard. Error: {error_str}",
                exc_info=True,
            )

            raise ValueError(error_str) from exc
        except requests.Timeout as exc:
            error_str = str(exc)

            logger.warning(msg="Connection to zKillboard timed out …")

            raise ValueError(error_str) from exc

        result_killmails = request_result.json()
        result = None
        for killmail in result_killmails:
            if killmail["killmail_id"] == int(kill_id):
                result = killmail

        if not result:
            logger.warning(
                msg=(
                    "Couldn't find any kill mail information in zKillboard's API response. "
                    "This is likely an issue with zKillboard."
                ),
                exc_info=True,
            )

            raise ValueError(
                "Couldn't find any kill mail information in zKillboard's API response. "
                "This is likely an issue with zKillboard."
            )

        try:
            killmail_id = result["killmail_id"]
            killmail_hash = result["zkb"]["hash"]

            esi_killmail = esi.client.Killmails.get_killmails_killmail_id_killmail_hash(
                killmail_id=killmail_id, killmail_hash=killmail_hash
            ).result()
        except Exception as exc:
            raise ValueError("Invalid Kill ID or Hash.") from exc

        # AA SRP
        from aasrp.models import Setting  # pylint: disable=import-outside-toplevel

        loss_value_field = Setting.objects.get_setting(Setting.Field.LOSS_VALUE_SOURCE)

        ship_type = esi_killmail["victim"]["ship_type_id"]
        ship_value = result["zkb"][loss_value_field]

        logger.debug(msg=f"Ship type for kill ID {kill_id} is {ship_type}")
        logger.debug(msg=f"Total loss value for kill id {kill_id} is {ship_value}")

        victim_id = esi_killmail["victim"]["character_id"]

        return ship_type, ship_value, victim_id

    @staticmethod
    def pending_requests_count_for_user(user: User):
        """
        Returns the number of open SRP requests for given user
        or None if user has no permission

        :param user:
        :type user:
        :return:
        :rtype:
        """

        # AA SRP
        from aasrp.models import SrpRequest  # pylint: disable=import-outside-toplevel

        if user.has_perm(perm="aasrp.manage_srp") or user.has_perm(
            perm="aasrp.manage_srp_requests"
        ):
            return SrpRequest.objects.filter(
                request_status=SrpRequest.Status.PENDING
            ).count()

        return None

    @staticmethod
    def get_insurance_for_ship_type(ship_type_id: int):
        """
        Getting insurance for a given ship type ID from ESI

        :param ship_type_id:
        :type ship_type_id:
        :return:
        :rtype:
        """

        insurance_prices = esi.client.Insurance.get_insurance_prices().result()

        for insurance in insurance_prices:
            if insurance["type_id"] == ship_type_id:
                return insurance

        return None


class SettingQuerySet(models.QuerySet):
    """
    SettingQuerySet
    """

    def delete(self):
        """
        Delete action

        Override:
            We don't allow deletion here, so we make sure the object
            is saved again and not deleted

        :return:
        :rtype:
        """

        return super().update()


class SettingManager(models.Manager):
    """
    SettingManager
    """

    def get_setting(self, setting_key: str) -> str:
        """
        Return the value for given setting key

        :param setting_key:
        :type setting_key:
        :return:
        :rtype:
        """

        return getattr(self.first(), setting_key)

    def get_queryset(self):
        """
        Get a Setting queryset

        :return:
        :rtype:
        """

        return SettingQuerySet(self.model)



# Django
from django.test import TestCase

# AA SRP
from aasrp.tests.utils import create_fleettype


class TestFleetType(TestCase):
    """
    Tests for FleetType model
    """

    def test_model_string_names(self):
        """
        Test model string names

        :return:
        :rtype:
        """

        topic = create_fleettype(name="Test Fleet Type")

        self.assertEqual(first=str(topic), second="Test Fleet Type")



"""
Versioned static URLs to break browser caches when changing the app version
"""

# Django
from django.contrib.auth.models import User
from django.template.defaulttags import register
from django.templatetags.static import static

# Alliance Auth
from allianceauth.framework.api.user import get_main_character_name_from_user

# AA SRP
from aasrp import __version__


@register.simple_tag
def aasrp_static(path: str) -> str:
    """
    Versioned static URL
    Adding the app version to any static file we load through this function.
    This is to make sure to break the browser cache on app updates.

    Example: /static/myapp/css/myapp.css?ver=1.0.0

    :param path:
    :type path:
    :return:
    :rtype:
    """

    static_url = static(path)
    versioned_url = static_url + "?v=" + __version__

    return versioned_url


@register.filter
def main_character_name(user: User) -> str:
    """
    Get the users main character name, or return the username if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    return get_main_character_name_from_user(user=user)


@register.filter
def main_character_id(user: User) -> int:
    """
    Get the users main character id, or return 1 if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    if user is None:
        return 1

    try:
        return_value = user.profile.main_character.character_id
    except AttributeError:
        return_value = 1

    return return_value


@register.filter
def main_character_corporation_name(user: User) -> str:
    """
    Get the users main character corporation name,
    or an empty string if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    if user is None:
        return ""

    try:
        return_value = user.profile.main_character.corporation_name
    except AttributeError:
        return_value = ""

    return return_value


@register.filter
def main_character_corporation_id(user: User) -> int:
    """
    Get the users main character corporation id, or 1 if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    if user is None:
        return 1

    try:
        return_value = user.profile.main_character.corporation_id
    except AttributeError:
        return_value = 1

    return return_value


@register.filter
def main_character_alliance_name(user: User) -> str:
    """
    Get the users main character alliance name, or an empty string if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    if user is None:
        return ""

    try:
        return_value = user.profile.main_character.alliance_name
    except AttributeError:
        return_value = ""

    return return_value


@register.filter
def main_character_alliance_id(user: User) -> int:
    """
    Get the users main character alliance id, or 1 if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    if user is None:
        return 1

    try:
        return_value = user.profile.main_character.alliance_id

        # Check if the user is in an alliance
        try:
            return_value = int(return_value)
        except Exception:  # pylint: disable=broad-exception-caught
            return_value = 1
    except AttributeError:
        return_value = 1

    return return_value
